#include <boost/reflection.hpp>
#include <boost/serializer.hpp>

#include <boost/bind.hpp>
#include <boost/function.hpp>
#include <boost/lexical_cast.hpp>

#include <algorithm>
#include <iostream>
#include <iterator>
#include <sstream>
#include <string>
#include <vector>

class little {
	private:
		static const std::string xml_begin;
		static const std::string xml_end;

		int value;
	public:
		BOOST_SERIALIZER_ACCESSIBLE(little,
			BOOST_SERIALIZER_CONST_MEMBER_VAR(std::string, xml_begin);
			BOOST_SERIALIZER_CONST_MEMBER_VAR(std::string, xml_end);
			BOOST_SERIALIZER_MEMBER_VAR(int, value);
		)
		BOOST_SERIALIZER(little, std::ostream, xml,
			(BOOST_SERIALIZER_ITEM(xml_begin, boost::serializer::insert),
			(BOOST_SERIALIZER_ITEM(value, boost::serializer::insert),
			(BOOST_SERIALIZER_ITEM(xml_end, boost::serializer::insert),
			BOOST_PP_NIL)))
		)

		void serialize_xml(std::ostream &ostr) const {
			boost::serializer::oserialize <BOOST_SERIALIZER_TYPE(xml)> oser(little::xml, ostr);

			oser << *this;
		}

		little(int value_) : value(value_) {
		}
		little(const little &other) : value(other.value) {
		}
};

const std::string little::xml_begin = "\t<little value=\"";
const std::string little::xml_end   = "\" />\n";

BOOST_SERIALIZER_DEFINE(little, xml,
	(BOOST_SERIALIZER_DEFINE_ITEM((), ()),
	(BOOST_SERIALIZER_DEFINE_ITEM((), ()),
	(BOOST_SERIALIZER_DEFINE_ITEM((), ()),
	BOOST_PP_NIL)))
)

class big {
	public:
		typedef std::vector <little> values_type;
	private:
		static const std::string xml_begin;
		static const std::string xml_end;

		values_type values;
	public:
		BOOST_SERIALIZER_ACCESSIBLE(big,
			BOOST_SERIALIZER_CONST_MEMBER_VAR(std::string, xml_begin);
			BOOST_SERIALIZER_CONST_MEMBER_VAR(std::string, xml_end);
			BOOST_SERIALIZER_MEMBER_VAR(values_type, values);
		)
		BOOST_SERIALIZER(big, std::ostream, xml,
			(BOOST_SERIALIZER_ITEM(xml_begin, boost::serializer::insert),
			(BOOST_SERIALIZER_CONTAINER(values, boost::serializer::custom<boost::function<void(std::ostream &, const values_type::value_type &)> >::type),
			(BOOST_SERIALIZER_ITEM(xml_end, boost::serializer::insert),
			BOOST_PP_NIL)))
		)

		void serialize_xml(std::ostream &ostr) const {
			boost::serializer::oserialize <BOOST_SERIALIZER_TYPE(xml)> oser(big::xml, ostr);

			oser << *this;
		}

		big(const values_type &values_) : values(values_) {
		}
};

const std::string big::xml_begin = "<big>\n";
const std::string big::xml_end   = "</big>\n";

BOOST_SERIALIZER_DEFINE(big, xml,
	(BOOST_SERIALIZER_DEFINE_ITEM     ((), ()),
	(BOOST_SERIALIZER_DEFINE_CONTAINER((), (boost::bind(&little::serialize_xml, _2, _1))),
	(BOOST_SERIALIZER_DEFINE_ITEM     ((), ()),
	BOOST_PP_NIL)))
)

int main() {
	std::ostringstream ostr;
	big::values_type values;

	values.push_back(little(0));
	values.push_back(little(1));
	values.push_back(little(2));
	values.push_back(little(3));

	{
		big alpha(values);

		alpha.serialize_xml(ostr);
	}

	std::cout << ostr.str();
}
