#ifndef BOOST_SERIALIZER_SERIALIZE
#define BOOST_SERIALIZER_SERIALIZE

#include <boost/serializer/straits_base.hpp>

namespace boost {

namespace serializer {

namespace detail {
	template <typename StraitsType, unsigned int Which, unsigned int Total>
	struct oserialize_inserter {
		static inline void go(const StraitsType &straits,
		                      typename StraitsType::receive_type &receive,
		                      const typename StraitsType::class_type &c) {
			boost::serializer::out<Which>(straits, receive, c);
			oserialize_inserter<StraitsType, Which+1, Total>::go(straits, receive, c);
		}
	};
	template <typename StraitsType, unsigned int Total>
	struct oserialize_inserter <StraitsType, Total, Total> {
		static inline void go(const StraitsType &straits,
		                      typename StraitsType::receive_type &receive,
		                      const typename StraitsType::class_type &c) {
			// Nothing
		}
	};
}

template <typename StraitsType>
class oserialize {
	private:
		typedef oserialize <StraitsType> this_type;
	public:
		typedef StraitsType straits_type;
		typedef typename straits_type::class_type class_type;
		typedef typename straits_type::receive_type receive_type;
	private:
		const straits_type &straits;
		receive_type       &receive;
	public:
		inline receive_type &recipient() {
			return receive;
		}

		this_type &operator<<(const class_type &c) {
			detail::oserialize_inserter<straits_type, 0, straits_type::static_size>::go(straits, receive, c);
			return *this;
		}

		oserialize(const straits_type &straits_, receive_type &receive_) :
			straits(straits_), receive(receive_) {
		}
};

}

}

#endif
