#ifndef BOOST_SERIALIZER_STRAITS_BASE
#define BOOST_SERIALIZER_STRAITS_BASE

#include <boost/config.hpp>
#include <boost/bind.hpp>
#include <boost/preprocessor.hpp>
#include <boost/type_traits.hpp>
#include <functional>

// Used by the specializations of serializer_traits
#include <boost/tuple/tuple.hpp>

namespace boost {

namespace serializer {

#define BOOST_SERIALIZER_MAX_ARGS 10

#define BOOST_SERIALIZER_APPEND2(a,b) a ## b
#define BOOST_SERIALIZER_APPEND(a,b) BOOST_SERIALIZER_APPEND2(a,b)

#define BOOST_SERIALIZE_ACCESSOR_GET(Name) BOOST_SERIALIZER_APPEND(get_, Name)

#define BOOST_SERIALIZE_TRAITS(Name) BOOST_SERIALIZER_APPEND(BOOST_SERIALIZER_APPEND(serializer__, Name), _type)

#define BOOST_SERIALIZE_TRAITS_TEMPLATE_GET(num)    BOOST_SERIALIZER_APPEND(Get, num)
#define BOOST_SERIALIZE_TRAITS_GET(num)             BOOST_SERIALIZER_APPEND(BOOST_SERIALIZER_APPEND(get, num), _type)
#define BOOST_SERIALIZE_TRAITS_TEMPLATE_EXPORT(num) BOOST_SERIALIZER_APPEND(Exp, num)
#define BOOST_SERIALIZE_TRAITS_EXPORT(num)          BOOST_SERIALIZER_APPEND(BOOST_SERIALIZER_APPEND(exp, num), _type)
#define BOOST_SERIALIZE_TRAITS_TEMPLATE_VALID(num)  BOOST_SERIALIZER_APPEND(Valid, num)
#define BOOST_SERIALIZE_TRAITS_VALID(num)           BOOST_SERIALIZER_APPEND(valid, num)

#define BOOST_SERIALIZE_TRAITS_TEMPLATE_ARG_DECLARE_DEFAULT(z, n, data) \
	typename BOOST_SERIALIZE_TRAITS_TEMPLATE_GET(n) = void, \
	typename BOOST_SERIALIZE_TRAITS_TEMPLATE_EXPORT(n) = void,
#define BOOST_SERIALIZE_TRAITS_TEMPLATE_ARG_DECLARE(z, n, data) \
	typename BOOST_SERIALIZE_TRAITS_TEMPLATE_GET(n), \
	typename BOOST_SERIALIZE_TRAITS_TEMPLATE_EXPORT(n),
#define BOOST_SERIALIZE_TRAITS_TEMPLATE_ARG_PASS(z, n, data) \
	BOOST_SERIALIZE_TRAITS_TEMPLATE_GET(n), \
	BOOST_SERIALIZE_TRAITS_TEMPLATE_EXPORT(n),

template <typename, typename>
struct null_export {
};

template <typename ClassType,
          typename RecvType,
          unsigned int Count,
          BOOST_PP_REPEAT(BOOST_SERIALIZER_MAX_ARGS, BOOST_SERIALIZE_TRAITS_TEMPLATE_ARG_DECLARE_DEFAULT, ())
          typename Dummy = void>
struct serialize_traits {
};

template <unsigned int Which, typename StraitsType>
inline void out(const StraitsType &straits, typename StraitsType::receive_type &receive, const typename StraitsType::class_type &c) {
	return boost::get<Which>(straits.export_fun)(receive, (boost::get<Which>(straits.get_fun)(c)));
}

#define BOOST_SERIALIZE_TRAITS_WRAPPER_TEMPLATE_ARG(z, n, data) \
	BOOST_SERIALIZE_TRAITS_TEMPLATE_ARG_DECLARE_DEFAULT(z, n, data) \
	bool BOOST_SERIALIZE_TRAITS_TEMPLATE_VALID(n) = false,
#define BOOST_SERIALIZE_TRAITS_WRAPPER_VALID_CONST(z, n, data) \
	static const unsigned int BOOST_SERIALIZE_TRAITS_VALID(n) = BOOST_SERIALIZE_TRAITS_TEMPLATE_VALID(n) ? 1 : 0;
#define BOOST_SERIALIZE_TRAITS_WRAPPER_VALID_PLUS(z, n, data) \
	BOOST_SERIALIZE_TRAITS_VALID(n) +

template <typename ClassType,
          typename RecvType,
          BOOST_PP_REPEAT(BOOST_SERIALIZER_MAX_ARGS, BOOST_SERIALIZE_TRAITS_WRAPPER_TEMPLATE_ARG, ())
          typename Dummy = void>
struct serialize_traits_wrapper {
	BOOST_PP_REPEAT(BOOST_SERIALIZER_MAX_ARGS, BOOST_SERIALIZE_TRAITS_WRAPPER_VALID_CONST, ())
	static const unsigned int count =
		BOOST_PP_REPEAT(BOOST_SERIALIZER_MAX_ARGS, BOOST_SERIALIZE_TRAITS_WRAPPER_VALID_PLUS, ())
		0;
	typedef serialize_traits <
		ClassType,
		RecvType,
		count,
		BOOST_PP_REPEAT(BOOST_SERIALIZER_MAX_ARGS, BOOST_SERIALIZE_TRAITS_TEMPLATE_ARG_PASS, ())
		void>
		type;
};

// Used for BOOST_SERIALIZER_ACCESSIBLE
#define BOOST_SERIALIZER_CONST_MEMBER_VAR(Type, MemVar) \
	struct BOOST_SERIALIZE_ACCESSOR_GET(MemVar) : \
		std::unary_function<boost::add_reference<boost::add_const< serialize_accessor__class_type >::type>::type, \
		                    boost::add_reference<boost::add_const< Type >::type>::type> { \
		typedef BOOST_SERIALIZE_ACCESSOR_GET(MemVar) this_type; \
		inline this_type::result_type operator()(this_type::argument_type c) const { \
			return c.MemVar; \
		} \
	};
#define BOOST_SERIALIZER_MEMBER_VAR(Type, MemVar) \
	struct BOOST_SERIALIZE_ACCESSOR_GET(MemVar) : \
		std::unary_function<boost::add_reference<boost::add_const< serialize_accessor__class_type >::type>::type, \
		                    boost::add_reference<boost::add_const< Type >::type>::type> { \
		typedef BOOST_SERIALIZE_ACCESSOR_GET(MemVar) this_type; \
		inline this_type::result_type operator()(this_type::argument_type c) const { \
			return c.MemVar; \
		} \
	};
#define BOOST_SERIALIZER_CONST_MEMBER_FUNC(Type, GetMemFunc, Args) \
	struct BOOST_SERIALIZE_ACCESSOR_GET(MemFunc) : \
		std::unary_function<boost::add_reference<boost::add_const< serialize_accessor__class_type >::type>::type, \
		                    boost::add_reference<boost::add_const< Type >::type>::type> { \
		typedef BOOST_SERIALIZE_ACCESSOR_GET(GetMemFunc) this_type; \
		inline this_type::result_type operator()(this_type::argument_type c) const { \
			return c.MemFunc Args; \
		} \
	};

#define BOOST_SERIALIZER_ACCESSIBLE(ClassType, ETC) \
	struct serialize_accessor { \
		typedef ClassType serialize_accessor__class_type; \
		ETC \
	};

// Used for BOOST_SERIALIZER's ItemList
#define BOOST_SERIALIZER_ITEM_MACRO(RecvType, Name, Item, Export, CastOp) \
	serialize_accessor::BOOST_SERIALIZE_ACCESSOR_GET(Item), \
	Export < \
		RecvType, \
		boost::remove_reference<serialize_accessor::BOOST_SERIALIZE_ACCESSOR_GET(Item)::result_type>::type, \
		CastOp \
	>,
#define BOOST_SERIALIZER_ITEM_MACRO_STUB(RecvType, Name, Params) \
	BOOST_SERIALIZER_ITEM_MACRO(RecvType, Name, \
		BOOST_PP_TUPLE_ELEM(3, 0, Params), \
		BOOST_PP_TUPLE_ELEM(3, 1, Params), \
		BOOST_PP_TUPLE_ELEM(3, 2, Params) \
	)
#define BOOST_SERIALIZER_ITEM_CAST(Item, Export, CastOp) \
	(BOOST_SERIALIZER_ITEM_MACRO_STUB, (Item, Export, CastOp))
#define BOOST_SERIALIZER_ITEM(Item, Export) \
	BOOST_SERIALIZER_ITEM_CAST(Item, Export, boost::serializer::nop<serialize_accessor::BOOST_SERIALIZE_ACCESSOR_GET(Item)::result_type>)

#define BOOST_SERIALIZER_CONTAINER_MACRO(RecvType, Name, Item, Export, CastOp) \
	serialize_accessor::BOOST_SERIALIZE_ACCESSOR_GET(Item), boost::serializer::for_each< \
		RecvType, \
		boost::remove_reference<serialize_accessor::BOOST_SERIALIZE_ACCESSOR_GET(Item)::result_type>::type, \
		Export, \
		CastOp \
	>,
#define BOOST_SERIALIZER_CONTAINER_CAST(Item, Export, CastOp) \
	(BOOST_SERIALIZER_CONTAINER_MACRO_STUB, (Item, Export, CastOp))
#define BOOST_SERIALIZER_CONTAINER_MACRO_STUB(RecvType, Name, Params) \
	BOOST_SERIALIZER_CONTAINER_MACRO(RecvType, Name, \
		BOOST_PP_TUPLE_ELEM(3, 0, Params), \
		BOOST_PP_TUPLE_ELEM(3, 1, Params), \
		BOOST_PP_TUPLE_ELEM(3, 2, Params) \
	)
#define BOOST_SERIALIZER_CONTAINER(Item, Export) \
	BOOST_SERIALIZER_CONTAINER_CAST( \
		Item, \
		Export, \
		boost::serializer::nop<boost::remove_reference<serialize_accessor::BOOST_SERIALIZE_ACCESSOR_GET(Item)::result_type>::type::value_type> \
	)

#define BOOST_SERIALIZER_ITEM_LIST_TO_TEMPLATE_PARAMS(r, data, elem) \
	BOOST_PP_TUPLE_ELEM(2, 0, elem)( \
		BOOST_PP_TUPLE_ELEM(3, 1, data), \
		BOOST_PP_TUPLE_ELEM(3, 2, data), \
		BOOST_PP_TUPLE_ELEM(2, 1, elem) \
	) true,

#define BOOST_SERIALIZER(ClassType, RecvType, Name, ItemList) \
	typedef boost::serializer::serialize_traits_wrapper <ClassType, RecvType, \
		BOOST_PP_LIST_FOR_EACH(BOOST_SERIALIZER_ITEM_LIST_TO_TEMPLATE_PARAMS, (ClassType, RecvType, Name), ItemList) \
	void>::type BOOST_SERIALIZE_TRAITS(Name); \
	static const BOOST_SERIALIZE_TRAITS(Name) Name;

// Used for BOOST_SERIALIZER_DEFINE ItemList
#define BOOST_SERIALIZER_DEFINE_ITEM_MACRO(ClassType, Name, Index, GetETC, ExportETC) \
	(0, ClassType::BOOST_SERIALIZE_TRAITS(Name)::BOOST_SERIALIZE_TRAITS_GET(Index)    GetETC), \
	(0, ClassType::BOOST_SERIALIZE_TRAITS(Name)::BOOST_SERIALIZE_TRAITS_EXPORT(Index) ExportETC),
#define BOOST_SERIALIZER_DEFINE_ITEM_MACRO_STUB(ClassType, Name, Index, Params) \
	BOOST_SERIALIZER_DEFINE_ITEM_MACRO(ClassType, Name, Index, \
		BOOST_PP_TUPLE_ELEM(2, 0, Params), \
		BOOST_PP_TUPLE_ELEM(2, 1, Params) \
	)
#define BOOST_SERIALIZER_DEFINE_ITEM(GetETC, ExportETC) \
	(BOOST_SERIALIZER_DEFINE_ITEM_MACRO_STUB, (GetETC, ExportETC))

#define BOOST_SERIALIZER_DEFINE_CONTAINER_MACRO(ClassType, Name, Index, GetETC, ExportETC) \
	(0, ClassType::BOOST_SERIALIZE_TRAITS(Name)::BOOST_SERIALIZE_TRAITS_GET(Index) GetETC), \
	(0, ClassType::BOOST_SERIALIZE_TRAITS(Name)::BOOST_SERIALIZE_TRAITS_EXPORT(Index)( \
		ClassType::BOOST_SERIALIZE_TRAITS(Name)::BOOST_SERIALIZE_TRAITS_EXPORT(Index)::pack_op_type ExportETC \
		)),
#define BOOST_SERIALIZER_DEFINE_CONTAINER_MACRO_STUB(ClassType, Name, Index, Params) \
	BOOST_SERIALIZER_DEFINE_CONTAINER_MACRO(ClassType, Name, Index, \
		BOOST_PP_TUPLE_ELEM(2, 0, Params), \
		BOOST_PP_TUPLE_ELEM(2, 1, Params) \
	)
#define BOOST_SERIALIZER_DEFINE_CONTAINER(GetETC, ExportETC) \
	(BOOST_SERIALIZER_DEFINE_CONTAINER_MACRO_STUB, (GetETC, ExportETC))

#define BOOST_SERIALIZER_DEFINE_ITEM_LIST_TO_PARAMS(r, data, i, elem) \
	BOOST_PP_TUPLE_ELEM(2, 0, elem)( \
		BOOST_PP_TUPLE_ELEM(2, 0, data), \
		BOOST_PP_TUPLE_ELEM(2, 1, data), \
		i, \
		BOOST_PP_TUPLE_ELEM(2, 1, elem) \
	)

#define BOOST_SERIALIZER_DEFINE(ClassType, Name, ItemList) \
	const ClassType::BOOST_SERIALIZE_TRAITS(Name) ClassType::Name( \
		BOOST_PP_LIST_FOR_EACH_I(BOOST_SERIALIZER_DEFINE_ITEM_LIST_TO_PARAMS, (ClassType, Name), ItemList) \
		true);

#define BOOST_SERIALIZER_TYPE(Name) \
	BOOST_SERIALIZE_TRAITS(Name)

// Used by the specializations of serializer_traits
#define BOOST_SERIALIZE_TRAITS_SPEC_TYPEDEFS(z, n, data) \
	typedef BOOST_SERIALIZE_TRAITS_TEMPLATE_GET(n) \
		BOOST_SERIALIZE_TRAITS_GET(n); \
	typedef BOOST_SERIALIZE_TRAITS_TEMPLATE_EXPORT(n) \
		BOOST_SERIALIZE_TRAITS_EXPORT(n);
#define BOOST_SERIALIZE_TRAITS_SPEC_GET_TYPES(z, n, data) \
	BOOST_SERIALIZE_TRAITS_GET(n),
#define BOOST_SERIALIZE_TRAITS_SPEC_EXPORT_TYPES(z, n, data) \
	BOOST_SERIALIZE_TRAITS_EXPORT(n),
#define BOOST_SERIALIZE_TRAITS_SPEC_CTOR_PARAMS(z, n, data) \
	boost::add_reference<boost::add_const<BOOST_SERIALIZE_TRAITS_GET(n)>::type>::type \
	BOOST_SERIALIZER_APPEND(g, n) \
	= BOOST_SERIALIZE_TRAITS_GET(n)(), \
	boost::add_reference<boost::add_const<BOOST_SERIALIZE_TRAITS_EXPORT(n)>::type>::type \
	BOOST_SERIALIZER_APPEND(e, n) \
	= BOOST_SERIALIZE_TRAITS_EXPORT(n)(),
#define BOOST_SERIALIZE_TRAITS_SPEC_CTOR_GET_INITLIST(z, n, data) \
	BOOST_SERIALIZER_APPEND(g, n),
#define BOOST_SERIALIZE_TRAITS_SPEC_CTOR_EXPORT_INITLIST(z, n, data) \
	BOOST_SERIALIZER_APPEND(e, n),

}

}

#endif
